/*
 * Copyright (c) 2003 David E. Bakken, his research students, and Washington State University.
 * Please see the file LICENSE.pdf for more details on terms of use.
 */

package ZEN_IDL.be;

import java.io.*;
import java.util.*;
import ZEN_IDL.ast.*;
import ZEN_IDL.be.helperUtils.*;
import util.*;

/*********************************************************
* Code generation helper. Takes care of mapping of IDL 
* types to sizes, java types, and java holder. Prints
* methods, parameter types, and parameter names.
*
* @author Olav Haugan
*
* 14.10.01
*********************************************************/
public class MQCCodeGenHelper
{
	private Hashtable IDLTypes; /* Map from IDL to Java */
	private Hashtable IDLRWMap; /* Map from IDL to read/write marshall/unmarshall functions */
	private Hashtable IDLTypesHolder; /* Map from IDL to holder */
	private Hashtable IDLSize; /* Map from IDL to sizes */
	private Hashtable alignment; /* Alignment for types */
	
	private MQCConfig config; //The configuration options

	public static HashMap exceptionMap = new HashMap();
		
	/*********************************************************
	* Constructor. Initializes variables.
	* Sets uo a map of IDLtypes -> java types, IDLtypes ->
	* Java-size, IDLtype -> alignment size,  IDLtype->Java type
	* holder.
	*********************************************************/
	public MQCCodeGenHelper(MQCConfig config)
	{
		this.config = config;
		
		//IDL to java mapping
		IDLTypes = new Hashtable(16);
		IDLTypes.put( "octet" , "byte" );
		IDLTypes.put( "boolean", "boolean" );
		IDLTypes.put( "char" , "char" );
		IDLTypes.put( "wchar", "char" );
		IDLTypes.put( "string", "String" );
		IDLTypes.put( "wstring", "String" );
		IDLTypes.put( "short", "short" );
		IDLTypes.put( "unsigned short", "short" );
		IDLTypes.put( "long", "int" );
		IDLTypes.put( "unsigned long", "int" );
		IDLTypes.put( "long long", "long" );
		IDLTypes.put( "unsigned long long", "long" );
		IDLTypes.put( "float", "float" );
		IDLTypes.put( "double", "double" );
		IDLTypes.put( "long double", "double" );
		IDLTypes.put( "void", "void" );

		//IDL to java type byte size mapping
		IDLSize = new Hashtable(16);
		IDLSize.put( "octet" , new Integer(1) );
		IDLSize.put( "boolean", new Integer(1) );
		IDLSize.put( "char" , new Integer(1) );
		IDLSize.put( "wchar", new Integer(1) );
		IDLSize.put( "string", new Integer(0) );
		IDLSize.put( "wstring", new Integer(0) );
		IDLSize.put( "short", new Integer(2) ); 
		IDLSize.put( "unsigned short", new Integer(2) );
		IDLSize.put( "long", new Integer(4) );
		IDLSize.put( "unsigned long", new Integer(4) );
		IDLSize.put( "long long", new Integer(8) );
		IDLSize.put( "unsigned long long", new Integer(8) );
		IDLSize.put( "float", new Integer(4) );
		IDLSize.put( "double", new Integer(8) );
//		IDLSize.put( "long double", new Integer(16) );
		IDLSize.put( "void", new Integer(0) );

		//IDL to java alignment mapping
		alignment = new Hashtable(17);
		alignment.put( "octet" , new Integer(1) );
		alignment.put( "boolean", new Integer(1) );
		alignment.put( "char" , new Integer(1) );
		alignment.put( "wchar", new Integer(1) );
		alignment.put( "short", new Integer(2) ); 
		alignment.put( "unsigned short", new Integer(2) );
		alignment.put( "long", new Integer(4) );
		alignment.put( "unsigned long", new Integer(4) );
		alignment.put( "float", new Integer(4) );
		alignment.put( "enum", new Integer(4) );
		alignment.put( "double", new Integer(8) );
//		alignment.put( "long double", new Integer(8) );
		alignment.put( "long long", new Integer(8) );
		alignment.put( "unsigned long long", new Integer(8) );
		alignment.put( "void", new Integer(0) );
		alignment.put( "string", new Integer(0) );
		alignment.put( "wstring", new Integer(0) );

		//IDL to java CDR read and write function names
		IDLRWMap = new Hashtable(15);
		IDLRWMap.put( "octet" , "octet" );
		IDLRWMap.put( "boolean", "boolean" );
		IDLRWMap.put( "char" , "char" );
		IDLRWMap.put( "wchar", "char" );
		IDLRWMap.put( "string", "string" );
		IDLRWMap.put( "wstring", "string" );
		IDLRWMap.put( "short", "short" );
		IDLRWMap.put( "unsigned short", "ushort" );
		IDLRWMap.put( "long", "long" );
		IDLRWMap.put( "unsigned long", "ulong" );
		IDLRWMap.put( "long long", "longlong" );
		IDLRWMap.put( "unsigned long long", "ulonglong" );
		IDLRWMap.put( "float", "float" );
		IDLRWMap.put( "double", "double" );
//		IDLRWMap.put( "long double", "double" );		

		IDLTypesHolder = new Hashtable(15);
		IDLTypesHolder.put( "octet" , "mqc.holders.ByteHolder" );
		IDLTypesHolder.put( "boolean", "mqc.holders.BooleanHolder" );
		IDLTypesHolder.put( "char" , "mqc.holders.CharHolder" );
		IDLTypesHolder.put( "wchar", "mqc.holders.CharHolder" );
		IDLTypesHolder.put( "string", "mqc.holders.StringHolder" );
		IDLTypesHolder.put( "wstring", "mqc.holders.StringHolder" );
		IDLTypesHolder.put( "short", "mqc.holders.ShortHolder" );
		IDLTypesHolder.put( "unsigned short", "mqc.holders.ShortHolder" );
		IDLTypesHolder.put( "long", "mqc.holders.IntHolder" );
		IDLTypesHolder.put( "unsigned long", "mqc.holders.IntHolder" );
		IDLTypesHolder.put( "long long", "mqc.holders.LongHolder" );
		IDLTypesHolder.put( "unsigned long long", "mqc.holders.LongHolder" );
		IDLTypesHolder.put( "float", "mqc.holders.FloatHolder" );
		IDLTypesHolder.put( "double", "mqc.holders.DoubleHolder" );


		//IDL to java type holder mapping
/*	
		IDLTypesHolder = new Hashtable(15);
		IDLTypesHolder.put( "octet" , "org.omg.CORBA.ByteHolder" );
		IDLTypesHolder.put( "boolean", "org.omg.CORBA.BooleanHolder" );
		IDLTypesHolder.put( "char" , "org.omg.CORBA.CharHolder" );
		IDLTypesHolder.put( "wchar", "org.omg.CORBA.CharHolder" );
		IDLTypesHolder.put( "string", "org.omg.CORBA.StringHolder" );
		IDLTypesHolder.put( "wstring", "org.omg.CORBA.StringHolder" );
		IDLTypesHolder.put( "short", "org.omg.CORBA.ShortHolder" );
		IDLTypesHolder.put( "unsigned short", "org.omg.CORBA.ShortHolder" );
		IDLTypesHolder.put( "long", "org.omg.CORBA.IntHolder" );
		IDLTypesHolder.put( "unsigned long", "org.omg.CORBA.IntHolder" );
		IDLTypesHolder.put( "long long", "org.omg.CORBA.LongHolder" );
		IDLTypesHolder.put( "unsigned long long", "org.omg.CORBA.LongHolder" );
		IDLTypesHolder.put( "float", "org.omg.CORBA.FloatHolder" );
		IDLTypesHolder.put( "double", "org.omg.CORBA.DoubleHolder" );
	*/
//		IDLTypesHolder.put( "long double", "org.omg.CORBA.DoubleHolder" );		
	}	
	
	/*********************************************************
	* Prints the method hearders
	* Ex. public long func(String arg);
	*
	* @param PrintWriter The stream to write to.
	*        
	* @param Vector The methods to print
	*********************************************************/
	public void printMethods( PrintWriter pw, Vector interfaceMethods )
	{
		Enumeration e = interfaceMethods.elements();
		InterfaceMethod iMethod = null;
		Vector params = null;
		
		while( e.hasMoreElements() ) 
		{
			iMethod = (InterfaceMethod) e.nextElement();
			params = iMethod.paramVector;
			pw.print("\tpublic " + mapIDL2Java( iMethod.returnType ) + " " + iMethod.name + "( " );
			printParams(pw, params);
			pw.println( " );");
		}
	}
	
	/*********************************************************
	* Prints the method parameters type and name.
	*
	* @param PrintWriter The stream to write to.
	*        
	* @param Vector The parameters to print
	*********************************************************/
	public void printParams( PrintWriter pw, Vector params )
	{
		Enumeration e = params.elements();
		InterfaceMethodParam arg = null;
		
		while( e.hasMoreElements() )
		{
			arg = (InterfaceMethodParam) e.nextElement();
			if( arg.paramInOut.equals("in") )
				pw.print( mapIDL2Java(arg.paramType) );
			else
				pw.print( mapIDL2JavaHolder(arg.paramType) );
			pw.print(" " + arg.paramName );
			if( e.hasMoreElements() )
				pw.print(", " );
		}
	}

	/*********************************************************
	* Prints the method parameter names
	*
	* @param PrintWriter The stream to write to.
	*        
	* @param Vector The parameter names to print
	*********************************************************/
	public void printParamNames(PrintWriter pw, Vector params)
	{
		Enumeration e = params.elements();
		InterfaceMethodParam arg = null;
		
		while( e.hasMoreElements() )
		{
			arg = (InterfaceMethodParam) e.nextElement();
			pw.print( arg.paramName );
			if( e.hasMoreElements() )
				pw.print(", " );
		}
	}
	
	/*********************************************************
	* Checks if a type is a primitive type or not.
	*
	* @param String The name of the type
	*        
	* @return boolean true if the type is primitive, false
	*          otherwise.
	*********************************************************/
	public boolean isPrimitiveType( String type )
	{
		return IDLTypes.containsKey( MQCTypedefMapping.get(type.trim()) );
	}
	
	/*********************************************************
	* Maps an IDL type to java
	*
	* @param String The name of the type
	*        
	* @return String The name of the java type
	*********************************************************/
	public String mapIDL2Java( String type )
	{
		String javaType = (String) IDLTypes.get( type.trim() );
		if(javaType == null)
		{
			String realType = MQCTypedefMapping.get(type);
			if(realType == type)
				return type;
			else
			{
				javaType = (String) IDLTypes.get( realType );
				return javaType;
			}
		}
		else
			return javaType;
	}

	/*********************************************************
	* Maps an IDL type to the name of the CDRStream read
	* or write function.
	*
	* @param String The name of the type
	*        
	* @return String The name of the read or write function
	*********************************************************/
	public String mapIDL2RWMap( String type )
	{
		String javaType = (String) IDLRWMap.get( type.trim() );
		if(javaType == null)
		{
			String realType = MQCTypedefMapping.get(type);
			if(realType == type)
				return type;
			else
			{
				javaType = (String) IDLRWMap.get( realType );
				return javaType;
			}
		}
		else
			return javaType;
	}
	
	/*********************************************************
	* Maps an IDL type to a java holder. Holder is used if
	* we are using "out" arguments.
	*
	* @param String The name of the type
	*        
	* @return String The name of the java holder
	*********************************************************/
	public String mapIDL2JavaHolder( String type )
	{
		String javaType = (String) IDLTypesHolder.get( type.trim() );
		if(javaType == null)
		{
			String realType = MQCTypedefMapping.get(type);
			if(realType == type)
				return (type + "Holder");
			else
			{
				javaType = (String) IDLTypesHolder.get( realType );
				return javaType;
			}
		}
		else
			return javaType;
	}

	/*********************************************************
	* Constructs a file and returns the IO stream writer
	* object.
	*
	* @param String The name of the package
	*        
	* @param String The name of file
	*        
	* @return PrintWriter The file stream generated
	*********************************************************/
	public static PrintWriter makeFile(String packagePath, String fileName)
	{
		return makeFileExt(packagePath, fileName, ".java");
	}

	/*********************************************************
	* Constructs a file and returns the IO stream writer object.
	*
	* @param String The name of the package
	* @param String The name of file
	* @return PrintWriter The file stream generated
	*********************************************************/
	public static PrintWriter makeFileExt(String packagePath, String fileName, String fileExt)
	{
		try
		{
			return new PrintWriter(new BufferedWriter(new FileWriter( packagePath.trim() + fileName.trim() + fileExt.trim() )), false);
		}
		catch(java.io.IOException e)
		{
			System.out.println("Could not generate file: " + packagePath.trim() + fileName.trim() + fileExt.trim());
			System.exit(-1);
		}
		return null;
	}

	/*********************************************************
	* Return the size of a java type
	* NB: Does not have support for complex types.
	*
	* @param String The name of the type
	*        
	* @return int The parameter size
	*********************************************************/
	public int getParamSize(String type)
	{
		String realType = MQCTypedefMapping.get(type);
		if(realType != type)
			type = realType;
		if(isPrimitiveType(type))
		{
			Object o = (IDLSize.get( type.trim() ));
			if(o == null)
				return 0;
			int size = 	(int) ((Integer)(o)).intValue();
			return size;
		}
		else
		{
			return 0;
		}
	}

	/*********************************************************
	* Gets the aligment for the java type
	* NB: Does not have support for complex types.
	*
	* @param String The name of the type
	*        
	* @return int The alignment
	*********************************************************/
	public int getAlignment(String type)
	{
		String realType = MQCTypedefMapping.get(type);
		if(realType != type)
			type = realType;

		return ((Integer) alignment.get( type.trim() )).intValue();
	}
	
	/*********************************************************
	* Return the size exact size of the in payload 
	* NB: Does not have support for complex types.
	*
	* @param InterfaceMethod The method
	*        
	* @param String The method name
	*        
	* @param int The total header size
	*        
	* @return int payload size
	*********************************************************/
	public int getInPayload(InterfaceMethod iMethod, String name, int hdrSize)
	{
		if(config.isPadding())
			return getInPayloadWithPadding(iMethod, name, hdrSize );
			
		Enumeration e = iMethod.paramVector.elements();
		InterfaceMethodParam arg = null;
		int size = hdrSize;
		size += name.length() + 1;
		
		while( e.hasMoreElements() )
		{
			arg = (InterfaceMethodParam) e.nextElement();
			if( !arg.paramInOut.equals("out") )
			{
				/* This must be changed so that we can do structs etc. etc */
				size += getParamSize(arg.paramType);
			}
		}
		return size - (hdrSize + name.length() + 1);
	}

	/*********************************************************
	* Return the size exact size of the payload with padding
	* NB: Does not have support for complex types.
	*
	* @param InterfaceMethod The method
	*        
	* @param String The method name
	*        
	* @param int The total header size
	*        
	* @return int payload size
	*********************************************************/
	public int getInPayloadWithPadding(InterfaceMethod iMethod, String name, int hdrSize)
	{
		Enumeration e = iMethod.paramVector.elements();
		InterfaceMethodParam arg = null;
		int size = hdrSize;
		size += name.length() + 1;

		while( e.hasMoreElements() )
		{
			arg = (InterfaceMethodParam) e.nextElement();
			if( !arg.paramInOut.equals("out") )
			{
				/* This must be changed so that we can do structs etc. etc */
				size = align(arg.paramType, size);
				size += getParamSize(arg.paramType);
			}
		}
		return size - (hdrSize + name.length() + 1);
	}
	
	/*********************************************************
	* Return the size exact size of the out payload 
	* NB: Does not have support for complex types.
	*
	* @param InterfaceMethod The method
	*        
	* @param String The method name
	*        
	* @param int The total header size
	*        
	* @return int payload size
	*********************************************************/
	public int getOutPayload(InterfaceMethod iMethod, int hdrSize)
	{
		if(config.isPadding())
			return getOutPayloadWithPadding(iMethod, hdrSize );
			
		Enumeration e = iMethod.paramVector.elements();
		InterfaceMethodParam arg = null;
		int size = hdrSize;
		
		if(!iMethod.returnType.equals("void"))
			size += getParamSize(iMethod.returnType);

		while( e.hasMoreElements() )
		{
			arg = (InterfaceMethodParam) e.nextElement();
			if( !arg.paramInOut.equals("out") )
			{
				/* This must be changed so that we can do structs etc. etc */
				size += getParamSize(arg.paramType);
			}
		}
		return size - hdrSize;
	}

	/*********************************************************
	* Return the size exact size of the payload with padding
	* NB: Does not have support for complex types.
	*
	* @param InterfaceMethod The method
	*        
	* @param int The total header size
	*        
	* @return int payload size
	*********************************************************/
	public int getOutPayloadWithPadding(InterfaceMethod iMethod, int hdrSize)
	{
		Enumeration e = iMethod.paramVector.elements();
		InterfaceMethodParam arg = null;
		int size = hdrSize;
		
		if(!iMethod.returnType.equals("void"))
		{
			size = align(iMethod.returnType, size);
			size += getParamSize(iMethod.returnType);
		}
		while( e.hasMoreElements() )
		{
			arg = (InterfaceMethodParam) e.nextElement();
			if( !arg.paramInOut.equals("in") )
			{
				/* This must be changed so that we can do structs etc. etc */
				size = align(arg.paramType, size);
				size += getParamSize(arg.paramType);
			}
		}
		return size - hdrSize;
	}

	/*********************************************************
	* Return the exact size of the exception payload
	* NB: Does not have support for complex types.
	*
	* @param int The size of the header
	*        
	* @param int the id length
	*        
	* @param String The name of the exception
	*        
	* @return int payload size
	*********************************************************/
	public int getExPayload(int hdrSize, int idLen, String name )
	{
		if(config.isPadding())
			return getExPayloadWithPadding(hdrSize, idLen, name );
					
		Vector ex = (Vector) exceptionMap.get(name);
		
		int size = 4 + idLen;
		Enumeration e = ex.elements();
		while(e.hasMoreElements())
		{
			String type = e.nextElement().toString();
			size += getParamSize(type);
		}
		return size;
	}
					
	/*********************************************************
	* Return the exact size of the exception payload with
	* padding.
	* NB: Does not have support for complex types.
	*
	* @param int The size of the header
	*        
	* @param int the id length
	*        
	* @param String The name of the exception
	*        
	* @return int payload size
	*********************************************************/
	public int getExPayloadWithPadding(int hdrSize, int idLen, String name )
	{
		int size = hdrSize;
		Vector ex = (Vector) exceptionMap.get(name);
		
		size = align("long", size);
		size += 4 + idLen;
		
		Enumeration e = ex.elements();
		while(e.hasMoreElements())
		{
			String type = e.nextElement().toString();
			size = align(type, size);
			size += getParamSize(type);
		}
		return size - hdrSize;
	}

	/*********************************************************
	* Return the size of the maximum payload 
	* NB: Does not have support for complex types.
	*
	* @param InterfaceMethod The method
	*        
	* @return int The maximum payload size
	*********************************************************/
	public int getPayload(InterfaceMethod iMethod)
	{
		Enumeration e = iMethod.paramVector.elements();
		InterfaceMethodParam arg = null;
		int inSize = 0;

		while( e.hasMoreElements() )
		{
			arg = (InterfaceMethodParam) e.nextElement();
			if( !arg.paramInOut.equals("out") )
			{
				/* This must be changed so that we can do structs etc. etc */
				inSize += getParamSize(arg.paramType);
			}
		}
		return inSize;
	}

	/*********************************************************
	* Computes the alignment
	* 
	* @param String The argument type
	*        
	* @param int The current index
	*
	* @return int The new index
	*********************************************************/
	public int align(String arg, int index)
	{
		int alignValue = getAlignment(arg);
		if(alignValue < 2)
			return index;
		index = (index + (alignValue-1)) & ~(alignValue-1);
		return index;
	}

	/*********************************************************
	* Generates the ID string for identifying interfaces
	* 
	* @param String The package path
	*        
	* @param String The name of the interface
	*
	* @return String The id string
	*********************************************************/
	public String generateID( String packagePath , String interfaceName)
	{
		StringTokenizer tok = new StringTokenizer( packagePath , File.separator );
		StringBuffer id = new StringBuffer("IDL:");
		while( tok.hasMoreTokens() )
		{
			id.append(tok.nextToken());
			id.append("/");
		}
		if(config.isCompressing())
		{
			id.append(MQCCompress.compressInterface(interfaceName));
		}
		else
		{
			id.append(interfaceName);
		}
		id.append(":1.0");
		return id.toString();
	}

	/*********************************************************
	* Generates the full name of objects
	* Converts from "/path/path2/etc" to "path.path2.etc" 
	*
	* @param String The package path
	*        
	* @param String The name of the object
	*
	* @return String The full name
	*********************************************************/
	public static String generateFullObjName( String packagePath , String name )
	{
		StringTokenizer tok = new StringTokenizer( packagePath , File.separator );
		StringBuffer full = new StringBuffer();
		while( tok.hasMoreTokens() )
		{
			full.append(tok.nextToken());
			full.append(".");
		}
		full.append(name);
		return full.toString();
	}

	/*********************************************************
	* Return the number of out and inout params.
	*
	* @param InterfaceMethod The method
	*        
	* @return int The number of out or inout params
	*********************************************************/
	public int noOutArgs(InterfaceMethod iMethod)
	{
		Enumeration e = iMethod.paramVector.elements();
		InterfaceMethodParam arg = null;

		int num = 0;
		if(!iMethod.returnType.equals("void"))
			num++;
		while( e.hasMoreElements() )
		{
			arg = (InterfaceMethodParam) e.nextElement();
			if( !arg.paramInOut.equals("in") )
			{
				num++;
			}
		}
		return num;
	}

}
