/*
 * Copyright (c) 2003 David E. Bakken, his research students, and Washington State University.
 * Please see the file LICENSE.pdf for more details on terms of use.
 */

package mqc.protocols;

/**
 * @author Kim Christian Swenson
 */
public abstract class GIOPHeader implements Marshalable
{
   /** First byte in a GIOP message is G */
   public final static byte GMAGIC = 0x47;

   /** Second byte in a GIOP message is I */
   public final static byte IMAGIC = 0x49;

   /** Third byte in a GIOP message is O */
   public final static byte OMAGIC = 0x4F;

   /** Fourth byte in a GIOP message is P */
   public final static byte PMAGIC = 0x50;


   protected byte m_msgType;

   protected int m_msgSize;

   protected boolean m_littleEndian;

   protected byte m_major;

   protected byte m_minor;


   /**
    * Gets the major version number of this GIOP header.
    *
    * @return  major version number
    */
   public int getMajor()
   {
      return m_major;
   }

   /**
    * Gets the minor version number of this GIOP header.
    *
    * @return  minor version number
    */
   public int getMinor()
   {
      return m_minor;
   }

   /**
    * @return  true if header uses Big endian marshaling
    */
   public boolean isBigEndian()
   {
      return ! m_littleEndian;
   }

   /**
    * Sets whether message uses Big endian marshaling.
    */
   public void setBigEndian( boolean bigEndian )
   {
      m_littleEndian = ! bigEndian;
   }

   /**
    * Gets the message type of this GIOP header.
    *
    * @return  message type
    */
   public byte getMessageType()
   {
      return m_msgType;
   }

   public void setMessageType( byte type )
   {
      m_msgType = type;
   }

   /**
    * Gets the message size of GIOP message, not including the header.
    *
    * @return  size of GIOP message, not including this header
    */
   public int getMessageSize()
   {
      return m_msgSize;
   }

   /**
    * Sets the message size of GIOP message, not including the header.
    *
    * @param size  size of message, not including GIOP header
    */
   public void setMessageSize( int size )
   {
      m_msgSize = size;
   }

   public static int getMsgSize( byte[] buf, boolean bigEndian )
   {
      return CDRInputStream.readInt( buf, 8, bigEndian );
   }

   public static void setMsgSize( byte[] buf, int value, boolean bigEndian )
   {
      CDROutputStream.writeInt( buf, 8, value, bigEndian );
   }

}
