/* Id: SharedSecret.java,v 1.5 2000/02/09 20:32:33 gelderen Exp
 *
 * Copyright (C) 1995-2000 The Cryptix Foundation Limited.
 * All rights reserved.
 *
 * Use, modification, copying and distribution of this software is subject
 * the terms and conditions of the Cryptix General Licence. You should have
 * received a copy of the Cryptix General Licence along with this library;
 * if not, you can download a copy from http://www.cryptix.org/ .
 */
package mqc.security.mac;


import mqc.security.InvalidKeyException;
import mqc.security.Key;
import mqc.security.Mac;


/**
 * SharedSecret MAC, a MAC with length 0.
 *
 * @author Jeroen C. van Gelderen (gelderen@cryptix.org)
 * @version Revision: 1.5 
 */
public final class SharedSecret extends Mac
{
   byte[] secret;

   public SharedSecret()
   {
   }

   /** Private constructor used for cloning */
   private SharedSecret(byte[] theSecret)
   {
      secret = new byte[theSecret.length];
      System.arraycopy(theSecret, 0, secret, 0, theSecret.length);
   }

   /**
    * Return the length of this Mac: 0
    */
   protected final int engineGetMacLength()
   {
      // our MAC is just the secret, so return it's length
      return secret.length;
   }

   protected final void engineInit(Key key)
   throws InvalidKeyException
   {
      if( !key.getFormat().equals("RAW") )
         throw new InvalidKeyException("SharedSecret accepts keys in 'RAW' format only.");

      // extract the keybytes
      secret = key.getEncoded();

      // reset the engine to initial state
      engineReset();
   }


   protected final void engineUpdate(byte input)
   {
      // nothing to do
   }


   protected final void engineUpdate(byte[] input, int offset, int len)
   {
      // nothing to do
   }


   protected final byte[] engineDoFinal()
   {
      byte[] sec = new byte[secret.length];
      System.arraycopy(secret, 0, sec, 0, secret.length);
      return sec;
   }


   protected final void engineReset()
   {
      // nothing to do
   }


   /**
    * Clone this MAC object.
    */
   public Object clone()
      throws CloneNotSupportedException
   {
      // create a new SharedSecret object with the current secret value
      return new SharedSecret(secret);
   }
}
