/*
 * Copyright (c) 2003 David E. Bakken, his research students, and Washington State University.
 * Please see the file LICENSE.pdf for more details on terms of use.
 */

include(`Config_macros.m4')

package mqc.transports;

import mqc.*;
import mqc.holders.*;
import mqc.ior.*;

import java.io.*;
import java.util.*;
__J2SE__(
import java.net.*;
import java.net.SocketOptions.*;
)
__CLDC__(
import javax.microedition.io.Connector;
import javax.microedition.io.StreamConnection;
import javax.microedition.io.StreamConnectionNotifier;
// Use the MemoryEfficientConnector (add at a later date)
//import com.ajile.microedition.io.MemoryEfficientConnector;
)

/**
* TCP transport class
*
* @author Olav Haugan
* @author David McKinnon
*/
public class S_TCPTransport extends S_Transport
{
   int av = -1;
   __J2SE__(
   private ServerSocket socket;
   )
   __CLDC__(
   private com.sun.cldc.io.j2me.serversocket.Protocol socket;
   )
   private TaggedProfile [] profile;

   /**
    * Constructor. Initializes variables.
    */
   public S_TCPTransport(S_ORB orb)
   {
      super(orb);
      socket = null;
      init();
   }

   /**
    * Initializes a connection. We'll connect to the server here.
    *
    * @return boolean true if socket created successfully, false otherwise
    */
   public boolean init()
   {

      String host;
      int port;
      __J2SE__(
      try
      {
         socket = new ServerSocket(0); //Finds a free port
         //## socket.setOption(SocketOptions.TCP_NODELAY, true);
         host = InetAddress.getLocalHost().getHostAddress();
         port = socket.getLocalPort();
      }
      catch(IOException e)
      {
         System.out.println("Error creating socket");
         System.out.println(e);
         return false;
      }
      )

      __CLDC__(
      port = 12345;
      String socketURL = ("serversocket://:" + port);
      try
      {
         socket = (com.sun.cldc.io.j2me.serversocket.Protocol) Connector.open(socketURL);
         // use the new MemoryEfficientConnector
         //socket = (com.sun.cldc.io.j2me.serversocket.Protocol) MemoryEfficientConnector.open("serversocket", 0, port);
      }
      catch (IOException ioe)
      {
         System.out.println("**Error could not connect");
         ioe.printStackTrace();
      }

      if (socket == null)
      {
         System.out.println("Problem opening server socket.");
         return false;
      }
      host = System.getProperty("localhost");
      if (host == null)
      {
         host = "localhost";
      }
      )

      profile = new TaggedProfile[Config.protocols.length];
      for (int i=0; i < profile.length; i++)
      {
         switch (Config.protocols[i])
         {
         __GIOP_PROTOCOL__(
         case Config.GIOP:
            profile[i] = new IIOPProfile();
            break;
         )
         __GIOPLITE_PROTOCOL__(
         case Config.GIOPLite:
            profile[i] = new IIOPLiteProfile();
            break;
         )
         __MQCIOP_PROTOCOL__(
         case Config.MQCIOP:
            profile[i] = new MQCIOPProfile();
            break;
         )
         default:
            throw new RuntimeException("No protocol specified");
         }

         profile[i].major = (byte) 1;
         __GIOP_PROTOCOL__(
         __GIOP10__(profile[i].minor = (byte) 0;)
         __GIOP11__(profile[i].minor = (byte) 1;)
         __GIOP12__(profile[i].minor = (byte) 2;)
         __GIOP13__(profile[i].minor = (byte) 3;))
         __GIOPLITE_PROTOCOL__(profile[i].minor = (byte) 1;)
         __MQCIOP_PROTOCOL__(profile[i].minor = (byte) 1;)
         profile[i].host  = host;
         profile[i].port  = port;
      }
      return true;
   }

   /**
    * Thread run function. Waits for connections
    */
   public void run()
   {
      while(S_ORB.keepRunning)
      {
         try
         {
            __J2SE__(
            __DEBUG3__( System.out.println("Calling socket.accept...");)
            Socket client = socket.accept();
            )
            __CLDC__(
            __DEBUG3__( System.out.println("Calling socket.acceptAndOpen...");)
            StreamConnection client = socket.acceptAndOpen();
            )
            new S_TCPHandler(client, orb);
         }
         catch(IOException e)
         {
            return;
         }
      }
      // hack to enforce the shutting down of the S_ORB
      if (!S_ORB.keepRunning)
      {
         S_ORB.shutDown(-1);
      }
   }

   /**
    * Returns the tagged profile
    *
    * @return TaggedProfile The tagged profile
    */
   public TaggedProfile [] getTaggedProfile()
   {
      return  profile;
   }
}
